// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v5.xx - Atari ST interlaced image reduction framework
//	1991-2012 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//
//		evil@DHS for noticing faults, making additional suggestions & testing ideas
//		Frank B for encouraging me back into Atari land
//		ppera for feedback and suggestions re: stream conversion for his CATA video project
//		Vincent Riviere's port of GCC 4.x.x for MiNT/TOS
//		various members of Atari-Forum for adding their 2p (you know who you are!)
//		*all* users of PhotoChrome, Apex etc. - past and future
// ---------------------------------------------------------------------------------------------------------------------

#ifndef config_h_
#define config_h_

// ====================================================================
// 	arguments / app configuration
// ====================================================================

// --------------------------------------------------------------------

class task
{
public:
	
	task(const char* const _infile, const char* const _outfile)
		: infile_(_infile), outfile_(_outfile)
	{
	}

	std::string infile_;
	std::string outfile_;		
};

// --------------------------------------------------------------------

enum LaceType
{
	LaceType_FIELD,
	LaceType_HORIZONTAL,
	LaceType_VERTICAL,
	LaceType_CHECKER,
	LaceType_PRNG,
	LaceType_SYNCHRONIZED,
	//
	LaceType_OFF,
	//
	LaceType__NUM_,
	//	
	LaceType_DEFAULT = LaceType_CHECKER
};

enum DitherType
{
	DitherType_OFF,
	DitherType_PRNG,
	DitherType_STIPPLE,
	DitherType_CPRNG,
	DitherType_TINY,
	DitherType_BAYER4,
	DitherType_MESHED,
	DitherType_PANMAGIC,
	DitherType_BAYER8,
	//
	DitherType__NUM_,
	//	
	DitherType_DEFAULT = DitherType_OFF
};

enum ErrorType
{
	ErrorType_OFF,
	ErrorType_FIELD,
	ErrorType_SPATIAL,
	//ErrorType_MULTIMODE,	// doesn't really make sense with recent changes
	//
	ErrorType__NUM_,
	//	
	ErrorType_DEFAULT = ErrorType_SPATIAL
};

enum SpatialType
{
	SpatialType_FLOYD,
	SpatialType_JAJUNI,
	SpatialType_ZHIGANG,
	SpatialType_SHIAU1,
	SpatialType_SHIAU2,
	SpatialType_STUCKI,
	SpatialType_BURKES,
	SpatialType_SIERRA,
	SpatialType_ATKINSON,
	//
	SpatialType__NUM_,
	//	
	SpatialType_DEFAULT = SpatialType_STUCKI
};

enum ReduceMethod
{
	ReduceMethod_SIMPLE,
	ReduceMethod_SCAN_SUBDIVISION,
	ReduceMethod_LCG,
	ReduceMethod_BIN_BALANCING_L,
	ReduceMethod_SCAN_SUBDIVISION2,
	ReduceMethod_SOM,
	ReduceMethod_BIN_BALANCING_H,
	//
	ReduceMethod__NUM_,
	//	
	ReduceMethod_DEFAULT = ReduceMethod_SCAN_SUBDIVISION2
};

enum ColourDepth
{
	ColourDepth_ST = 3,
	ColourDepth_STE = 4,
	ColourDepth_STE_X = 5,
	ColourDepth_FALCON = 6,
	//	
	ColourDepth_DEFAULT = ColourDepth_STE
};

enum FieldMode
{
	FieldMode_NoSharing,
	FieldMode_SharedBitmap,
	FieldMode_SharedPalette,
	//
	FieldMode__NUM_,
	//	
	FieldMode_DEFAULT = FieldMode_NoSharing
};

enum StorageFmt
{
	StorageFmt_PPRLE,				// plane-by-plane RLE (i.e. v3,v4 compatible)
	StorageFmt_UNCOMPRESSED,		// plane-by-plane without compression (i.e. v3,v4 compatible)
	StorageFmt_DIRECT_DISPLAY,		// uncompressed, raw display format (no reformatting/decompression required to display)
	StorageFmt_RAW_DISPLAY,			// uncompressed, raw display format minus PCS header
	StorageFmt_COMPRESSIBLE,		// uncompressed, but reorganised as nibbles/r/g/b-streams for better 3rd party compression e.g. huffman/lzw.. whatever
	//
	StorageFmt__NUM_,
	//	
	StorageFmt_DEFAULT = StorageFmt_PPRLE
};


// --------------------------------------------------------------------

class Arguments
{
public:
	
	Arguments()
		: verbose(false),
		  quiet(false),
		  diagnostics(false),
		  write_file(false),
		  waitkey(false),
		  //
		  xsize(320), ysize(200),
		  fields(2),
		  colours(16),
		  palettes(3),
		  palette_skip(2),
		  firstline(1),
		  colourdepth(ColourDepth_DEFAULT),
		  ditherlevel(1.0f),
		  errorlevel(0.999f),
		  erroriters(1),
		  pfm_gamma(1.0f),
		  pfm_influence(0.99999f),
		  nn_depth(10),
		  nn_breadth(0.1f),
		  nn_rate(0.0025f),
		  //
		  method(ReduceMethod_DEFAULT),
		  dithertype(DitherType_DEFAULT),
		  lacetype(LaceType_DEFAULT),
		  errortype(ErrorType_DEFAULT),
		  storage(StorageFmt_DEFAULT),
		  spatialtype(SpatialType_DEFAULT),
		  fieldmode(FieldMode_DEFAULT),
		  //
		  uPCT(0), aPCT(0),
		  uDPT(0), aDPT(0)
	{
	}
	
	~Arguments()
	{
		delete[] uPCT;
		delete[] uDPT;
	}

	// flags
	bool verbose;
	bool quiet;
	bool diagnostics;
	bool write_file;
	bool waitkey;

	// variables
	int colourdepth;	// default value by enum, but really a variable
	int xsize;
	int ysize;
	int fields;
	int colours;
	int palettes;
	int palette_skip;
	int firstline;
	float ditherlevel;
	float errorlevel;
	int erroriters;
	float pfm_gamma;
	float pfm_influence;
	int nn_depth;
	float nn_breadth;
	float nn_rate;
	
	// modes
	ReduceMethod method;
	LaceType lacetype;
	DitherType dithertype;
	ErrorType errortype;
	StorageFmt storage;
	SpatialType spatialtype;
	FieldMode fieldmode;
	
	// files
	std::string pfm_source;
	std::string pct_source;
	std::string dpt_source;

	// config-sourced buffers
	byte_t *uPCT;
	byte_t *aPCT;
	word_t *uDPT;
	word_t *aDPT;
	
	// conversion tasks
	std::vector<task> tasks;
};

// --------------------------------------------------------------------

#endif // config_h_
