// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v5.xx - Atari ST interlaced image reduction framework
//	1991-2012 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//
//		evil@DHS for noticing faults, making additional suggestions & testing ideas
//		Frank B for encouraging me back into Atari land
//		ppera for feedback and suggestions re: stream conversion for his CATA video project
//		Vincent Riviere's port of GCC 4.x.x for MiNT/TOS
//		various members of Atari-Forum for adding their 2p (you know who you are!)
//		*all* users of PhotoChrome, Apex etc. - past and future
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
// 	harness to process/convert a single file
// ====================================================================
//	note: this bit is volatile - likely to be refactored
// ====================================================================

#include <iostream>
#include <algorithm>
#include <string>

#include "common.h"
#include "config.h"

#include "FreeImage.h"

#include "imagepkg.h"
#include "pcsimage_std_sgl.h"
#include "pcsimage_std_dual.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------

bool diagnostics_ = false;

bool process_file(const char* const _name, const char* const _outname)
{
	// nudge image package into emitting version info before any other messages
	ImagePkg::instance();
	
	// create source image, and perform any preparation processing (i.e. bitdepth, size)
	
	bool success = true;

	PCSImage* img = 0;
	if (g_allargs.fields == 2)
		img = new PCSImage_std_dual(_name, g_allargs.xsize, g_allargs.ysize, g_allargs.colourdepth, g_allargs.palettes, g_allargs.palette_skip);
	else
		img = new PCSImage_std_sgl(_name, g_allargs.xsize, g_allargs.ysize, g_allargs.colourdepth, g_allargs.palettes, g_allargs.palette_skip);
			
	if (!img) { printf("error - could not allocate memory for image class!\n"); getchar(); exit(1); }

	std::string sname(_name);
	std::transform(sname.begin(), sname.end(), sname.begin(), mytolower);
	
	int type = -1;
	if (sname.find(".jpg") != std::string::npos)
		type = 0;
	else if (sname.find(".tga") != std::string::npos)
		type = 1;
	else if (sname.find(".png") != std::string::npos)
		type = 2;
	else
	{
	    if (sname.find("_test") == 0 && 
	    	sname.find(".") == std::string::npos)
		{
			std::string num = sname.substr(5, std::string::npos);
			type = atoi(num.c_str());
			printf("selected built-in test source (%d)...\n", type);
			type = -(type+1);
		}
		else
		{			
			success = false;
			printf("error - unknown source image file extension!\n");		
			//printf("press any key...\n"); getchar();
		}
	}

	if (1 && success)
	{	
		if (!g_allargs.quiet)
			printf("> reading source image [%s]...\n", _name);		

		if (!img->read_source(type))
		{
			success = false;
			printf("error reading source image!\n");	
			//printf("press any key...\n"); getchar();							
		}
		else
		{
			if (g_allargs.verbose)
				printf("> source image read ok.\n");
		}
	}

	if (1 && success)
	{	
		if (g_allargs.verbose)
			printf("> sanitising source image for conversion...\n");		

		if (!img->sanitise_source())
		{
			success = false;
			printf("error reading source image!\n");		
			//printf("press any key...\n"); getchar();							
		}
		else
		{
			if (g_allargs.verbose)
				printf("> image sanitised ok.\n");
		}
	}
			
	if (1 && success)
	{	
		// diagnostics: write reference image as PNG
		if (g_allargs.diagnostics)
		{
			if (g_allargs.verbose)
				printf("> diagnostics: writing reference image...\n");
				
			//std::string refname(_name);
			//size_t p = refname.find_last_of(".");
			//if (p != std::string::npos)
			//	refname = refname.substr(0, p);
			//refname.append("R.png");
			std::string refname("srcref.png");
			
			if (!img->write_reference(refname.c_str()))
			{
				success = false;
				printf("error writing reference PNG!\n");
				//printf("press any key...\n"); getchar();							
			}
			else
			{
				if (g_allargs.verbose)				
					printf("> diagnostics: reference PNG written ok.\n");
			}
		}
	}

#ifdef CRYPTOCHROME
	if (1 && success)
	{
		if (!g_allargs.quiet)
			printf("> converting image [cryptochrome]...\n");
			
		img->convert_cc();
	}
#endif

	// convert image...			
	if (1 && success)
	{
		if (!g_allargs.quiet)
			printf("> converting image...\n");
			
		if (!img->convert())
		{
			success = false;
			printf("> error converting image!\n");					
			//printf("press any key...\n"); getchar();							
		}
		else
		{
			if (g_allargs.verbose)
				printf("> image converted ok.\n");
		}
	}
	
	if (1 && success)
	{
		if (!g_allargs.quiet)
			printf("> creating output file...\n");
			
		std::string outname(_outname);
		if (outname.length() == 0)
		{
			outname = _name;
			size_t p = outname.find_last_of(".");
			if (p != std::string::npos)
				outname = outname.substr(0, p);
				
			if (g_allargs.storage == StorageFmt_RAW_DISPLAY)
				outname.append(".pcr");
			else
				outname.append(".pcs");
		}
						
		if (!img->generate_pcs_file(outname.c_str()))
		{
			success = false;
			printf("> error generating file format!\n");					
			//printf("press any key...\n"); getchar();							
		}
		else
		{			
			if (g_allargs.verbose)
				printf("> conversion completed ok.\n");
		}
	}
//#endif
	
	delete img; img = 0;
	
	// clean up image package
	ImagePkg::destroy();
		
	return success;
}