// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v5.xx - Atari ST interlaced image reduction framework
//	1991-2012 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//
//		evil@DHS for noticing faults, making additional suggestions & testing ideas
//		Frank B for encouraging me back into Atari land
//		ppera for feedback and suggestions re: stream conversion for his CATA video project
//		Vincent Riviere's port of GCC 4.x.x for MiNT/TOS
//		various members of Atari-Forum for adding their 2p (you know who you are!)
//		*all* users of PhotoChrome, Apex etc. - past and future
// ---------------------------------------------------------------------------------------------------------------------

// ====================================================================
//	reusable image processing stuff
// ====================================================================

#include "common.h"
#include "config.h"

#include "FreeImage.h"

#include "imagepkg.h"
#include "pcsimage.h"

extern Arguments g_allargs;

// --------------------------------------------------------------------

static const unsigned long LCGSEED = 0xA1794529 + (1723 * 1);

InterlaceUnit::InterlaceUnit(signed short _field, signed short _colourdepth, signed short _fields)
	: LCG(LCGSEED),	// interlace unit should be field-stable!
	  field_(_field),
	  colourdepth_(_colourdepth),
	  fields_(_fields)
{
	// begin with 1/2 of the native colour LSB
	int n = (8 - _colourdepth) - 2;
	interlace_offset_ = 1 << n;
}

void InterlaceUnit::reset()
{
	LCG::reset(LCGSEED);
	onReset();
}

// --------------------------------------------------------------------

class InterlaceUnit_NONE : public InterlaceUnit
{
public:

	InterlaceUnit_NONE(signed short _field, signed short _colourdepth, signed short _fields)
		: InterlaceUnit(_field, _colourdepth, _fields)
	{
	}
	
	virtual signed short sample(signed short _x, signed short _y)
	{
		signed short eval = (0 + ((0 + field_) & 1)) & 1;

		signed short interlace;
		if (eval)
			interlace = interlace_offset_+1;	
		else
			interlace = -interlace_offset_-1;
			
		return interlace;		 
	}

	virtual float fsample(signed short _x, signed short _y)
	{
		signed short eval = (0 + ((0 + field_) & 1)) & 1;

		float interlace = 0.5f / (float)nvi_crange(colourdepth_, fields_);

		if (eval)
			interlace = -interlace;
			
		return interlace;		 
	}

};

class InterlaceUnit_HORIZONTAL : public InterlaceUnit
{
public:

	InterlaceUnit_HORIZONTAL(signed short _field, signed short _colourdepth, signed short _fields)
		: InterlaceUnit(_field, _colourdepth, _fields)
	{
	}
	
	virtual signed short sample(signed short _x, signed short _y)
	{
		signed short interlace = 0;
		signed short eval = (0 + ((_y + field_) & 1)) & 1;
		
		if (eval)
			interlace = interlace_offset_;	
		return interlace;
	}

	virtual float fsample(signed short _x, signed short _y)
	{
		signed short eval = (0 + ((_y + field_) & 1)) & 1;

		float interlace = 0.5f / (float)nvi_crange(colourdepth_, fields_);

		if (eval)
			interlace = -interlace;
			
		return interlace;		 
	}
};

class InterlaceUnit_VERTICAL : public InterlaceUnit
{
public:

	InterlaceUnit_VERTICAL(signed short _field, signed short _colourdepth, signed short _fields)
		: InterlaceUnit(_field, _colourdepth, _fields)
	{
	}
	
	virtual signed short sample(signed short _x, signed short _y)
	{
		signed short interlace = 0;
		signed short eval = (_x + ((0 + field_) & 1)) & 1;
		if (eval)
			interlace = interlace_offset_;	
		return interlace;
	}
	
	virtual float fsample(signed short _x, signed short _y)
	{
		signed short eval = (_x + ((0 + field_) & 1)) & 1;

		float interlace = 0.5f / (float)nvi_crange(colourdepth_, fields_);

		if (eval)
			interlace = -interlace;
			
		return interlace;		 
	}	
};

class InterlaceUnit_CHECKER : public InterlaceUnit
{
public:

	InterlaceUnit_CHECKER(signed short _field, signed short _colourdepth, signed short _fields)
		: InterlaceUnit(_field, _colourdepth, _fields)
	{
	}
	
	virtual signed short sample(signed short _x, signed short _y)
	{
		signed short interlace = 0;
		signed short eval = (_x + ((_y + field_) & 1)) & 1;
		if (eval)
			interlace = interlace_offset_;	
		return interlace;
	}
	
	virtual float fsample(signed short _x, signed short _y)
	{
		signed short eval = (_x + ((_y + field_) & 1)) & 1;

		float interlace = 0.5f / (float)nvi_crange(colourdepth_, fields_);

		if (eval)
			interlace = -interlace;
			
		return interlace;		 
	}		
};

class InterlaceUnit_PRNG : public InterlaceUnit
{
public:

	InterlaceUnit_PRNG(signed short _field, signed short _colourdepth, signed short _fields)
		: InterlaceUnit(_field, _colourdepth, _fields)
	{
	}
	
	virtual signed short sample(signed short _x, signed short _y)
	{
		unsigned short eval = ((lcg_sample() >> 6) + field_) & 1;
		signed short interlace = 0;
		if (eval)
			interlace = interlace_offset_;	
		return interlace;
	}
	
	virtual float fsample(signed short _x, signed short _y)
	{
		unsigned short eval = ((lcg_sample() >> 6) + field_) & 1;

		float interlace = 0.5f / (float)nvi_crange(colourdepth_, fields_);

		if (eval)
			interlace = -interlace;
			
		return interlace;		 
	}	
};
	
// --------------------------------------------------------------------

void PCSImage::init_interlace_unit(int _field)
{
	delete laceunit_; laceunit_ = 0;
	
	InterlaceUnit* newunit = 0;	
	switch (g_allargs.lacetype)
	{
	case LaceType_OFF:
		break;

	default:
	case LaceType_SYNCHRONIZED:	// todo
	case LaceType_FIELD:
		newunit = new InterlaceUnit_NONE(_field, colourdepth_, fields_);
		break;
	case LaceType_HORIZONTAL:
		newunit = new InterlaceUnit_HORIZONTAL(_field, colourdepth_, fields_);
		break;
	case LaceType_VERTICAL:
		newunit = new InterlaceUnit_VERTICAL(_field, colourdepth_, fields_);
		break;
	case LaceType_CHECKER:
		newunit = new InterlaceUnit_CHECKER(_field, colourdepth_, fields_);
		break;
	case LaceType_PRNG:
		newunit = new InterlaceUnit_PRNG(_field, colourdepth_, fields_);
		break;
	}
	
	laceunit_ = newunit;	
}

// --------------------------------------------------------------------
