// ---------------------------------------------------------------------------------------------------------------------
//	PhotoChrome v5.xx - Atari ST interlaced image reduction framework
//	1991-2012 Douglas Little (d.m.l)
// ---------------------------------------------------------------------------------------------------------------------
//	special thanks to:
//
//		evil@DHS for noticing faults, making additional suggestions & testing ideas
//		Frank B for encouraging me back into Atari land
//		ppera for feedback and suggestions re: stream conversion for his CATA video project
//		Vincent Riviere's port of GCC 4.x.x for MiNT/TOS
//		various members of Atari-Forum for adding their 2p (you know who you are!)
//		*all* users of PhotoChrome, Apex etc. - past and future
// ---------------------------------------------------------------------------------------------------------------------
//
// some notes:
//
//	This source is freeware	(except the 3rd party material of course) - you can do what you like with it - but using it
//	or adapting it without stating where it came from... is always bad form. ;-)
//
//	I've done my best to improve the convertor as much as possible without breaking compatibility with the old PCS
//	file format and viewers. This has been a bit limiting. I will however be breaking compatibility in the future to
//	get better quality images out.
//
//	The source includes some bits and pieces from the 'bit-twiddling-hacks' and 'magic algorithms' material easily found
//	on the interwebs. I need to mention those because I didn't invent them.
//
//	CIE/RGB colourspace conversion implemented from pseudocode found at http://www.easyrgb.com
//
// 	Most of the main code was written from scratch over the course of about a week, in very limited spare time,
//	so if see something you'd have done very differently, keep this in mind.
//	
//	The code is still in flux - some bits still need refactored into units/handlers and better isolated into modules
//	before branching off your own version, consider what the next update might look like!
//
//	There are some hacks/concessions to get this to build under MiNT/GCC headers - some of them more or less bad form,
//	but in most cases easy to correct with a little effort. These will likely get fixed when I run out of more interesting
//	things to do elsewhere.
//
//	FreeImage provides lots of file formats for free - but it produces a huge binary even if very little is actually
//	used by the application. I have tried to exclude plugins from FreeImage when built with -DPLUGINS_LITE, but they 
//	have done their best to make excluding 'buitin' plugins (of which there are a great many) without breaking lots of 
//	things. it works for a few formats but not others - needs more work.
//	FreeImage has been included pre-compiled as a .a because I'm not quite ready to release my build patches for it...
//	...you can of course build your own though. expect your binary to grow by 4mb without modifications :-)
//
//
//	todo:
//
//	- cie space needs retried
//	- perception map should take linearity and sensitivity of colour channels into account
//	- error diagnostics should take perception map into account
//	- xxxunits should clone host ptr instead of passing lots of fields from PCSImage
//
// ====================================================================
// 	application entrypoint & commandline handling
// ====================================================================

#include <iostream>
#include <algorithm>
#include <string>
#include <vector>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <locale.h>

// --------------------------------------------------------------------

#include "common.h"	// must come before SimpleOpt/SimpleGlob
#include "config.h"
#include "csvreader.h"

// --------------------------------------------------------------------

const char* const version_string = "PhotoChrome ST/e image renderer v5.12 R2";

extern "C" unsigned char PCT_PCS[];				// default sync table
extern "C" unsigned short DPT_PCS[];			// default dynamic point / colour mask table
extern "C" unsigned short DPT_PCS_original[];

// --------------------------------------------------------------------
// portability stuff

// this stuff is needed for SimpleOpt/SimpleGlob
#if defined(_MSC_VER)
# include <windows.h>
# include <tchar.h>
#else
# define TCHAR		char
# define _T(x)		x
#endif

// --------------------------------------------------------------------
// 3rd party stuff

#include "SimpleOpt.h"
#include "SimpleGlob.h"

// --------------------------------------------------------------------

// define the ID values to indentify the option
enum
{ 
	OPT_HELP, 
	OPT_VERBOSE, 
	OPT_QUIET, 
	OPT_WAIT_KEY, 
	OPT_DIAGNOSTICS, 
	//
	OPT_STORAGE, 
	OPT_FIELDS, 
	OPT_FIELDMODE, 
	OPT_COL_DEPTH, 
	OPT_METHOD, 
	OPT_LACE_TYPE, 
	OPT_DITHER_TYPE, 
	OPT_DITHER_LEVEL, 
	OPT_ERROR_TYPE, 
	OPT_ERROR_LEVEL, 
	OPT_ERROR_ITERS, 
	OPT_SPATIAL_TYPE, 
	OPT_NN_DEPTH, 
	OPT_NN_BREADTH, 
	OPT_NN_RATE, 
	OPT_PFM_GAMMA, 
	OPT_PFM_INFLUENCE, 
	OPT_PFM_SOURCE, 
	OPT_PCT_SOURCE, 
	OPT_DPT_SOURCE, 
	OPT_XSIZE, 
	OPT_YSIZE, 
	OPT_PLANES, 
	OPT_MIN_COLOUR,
	OPT_MAX_COLOUR,
	OPT_RESERVE_COLOUR,
	OPT_PALETTE_SIZE,
	OPT_PALETTE_STRIDE,
	OPT_PALETTE_PRELOAD,
	//
	OPT_OUTFILE,
	OPT_INFILE,
};

CSimpleOpt::SOption g_Options[] = 
{
	{ OPT_HELP, 		_T("-?"),     			SO_NONE		}, // "-?"
	{ OPT_HELP, 		_T("-h"),     			SO_NONE		}, // "-h"
	{ OPT_HELP, 		_T("--help"), 			SO_NONE		}, // "--help"
	//
	{ OPT_VERBOSE, 		_T("-v"),     			SO_NONE		}, // "-v"
	{ OPT_VERBOSE,		_T("--verbose"), 		SO_NONE		}, // "--verbose"
	//
	{ OPT_QUIET, 		_T("-q"),     			SO_NONE		}, // "-q"
	{ OPT_QUIET,		_T("--quiet"), 			SO_NONE		}, // "--quiet"
	//
	{ OPT_WAIT_KEY,		_T("-k"),     			SO_NONE		}, // "-k"
	{ OPT_WAIT_KEY,		_T("--wait-key"), 		SO_NONE		}, // "--wait-key"
	//
	{ OPT_DIAGNOSTICS,	_T("-dg"),     			SO_NONE		}, // "-d"
	{ OPT_DIAGNOSTICS,	_T("--diagnostics"),    SO_NONE		}, // "--diagnostics"
	//
	{ OPT_STORAGE,		_T("-s"),     			SO_REQ_SEP	}, // "-s <formatcode>"
	{ OPT_STORAGE,		_T("--storage"),   		SO_REQ_SEP	}, // "--storage <formatcode>"
	//
	{ OPT_FIELDS,		_T("-f"),     			SO_REQ_SEP	}, // "-f <#fields=1 or 2>"
	{ OPT_FIELDS,		_T("--fields"),   		SO_REQ_SEP	}, // "--fields <#fields=1 or 2>"
	//
	{ OPT_FIELDMODE,	_T("-fm"),     			SO_REQ_SEP	}, // "-fm <0,1,2>"
	{ OPT_FIELDMODE,	_T("--fieldmode"),   	SO_REQ_SEP	}, // "--fieldmode <0,1,2>"
	//
	{ OPT_COL_DEPTH,	_T("-cd"),     			SO_REQ_SEP	}, // "-cd <#bits per channel>"
	{ OPT_COL_DEPTH,	_T("--colour-depth"),   SO_REQ_SEP	}, // "--colour-depth <#bits per channel>"
	//
	{ OPT_METHOD,		_T("-m"),     			SO_REQ_SEP	}, // "-m <method#>"
	{ OPT_METHOD,		_T("--method"),   		SO_REQ_SEP	}, // "--method <method#>"
	//
	{ OPT_LACE_TYPE,	_T("-lt"),     			SO_REQ_SEP	}, // "-lt <type#>"
	{ OPT_LACE_TYPE,	_T("--lace-type"),   	SO_REQ_SEP	}, // "--lace-type <type#>"
	//
	{ OPT_DITHER_TYPE,	_T("-dt"),     			SO_REQ_SEP	}, // "-dt <type#>"
	{ OPT_DITHER_TYPE,	_T("--dither-type"),   	SO_REQ_SEP	}, // "--dither-type <type#>"
	//
	{ OPT_DITHER_LEVEL,	_T("-dl"),     			SO_REQ_SEP	}, // "-dl <level>"
	{ OPT_DITHER_LEVEL,	_T("--dither-level"),   SO_REQ_SEP	}, // "--dither-level <level>"
	//
	{ OPT_ERROR_TYPE,	_T("-et"),     			SO_REQ_SEP	}, // "-et <type#>"
	{ OPT_ERROR_TYPE,	_T("--error-type"),   	SO_REQ_SEP	}, // "--error-type <type#>"
	//
	{ OPT_ERROR_LEVEL,	_T("-el"),     			SO_REQ_SEP	}, // "-el <0-100%>"
	{ OPT_ERROR_LEVEL,	_T("--error-level"),   	SO_REQ_SEP	}, // "--error-level <0-100%>"
	//
	{ OPT_ERROR_ITERS,	_T("-ei"),     			SO_REQ_SEP	}, // "-ei <#iterations>"
	{ OPT_ERROR_ITERS,	_T("--error-iters"),   	SO_REQ_SEP	}, // "--error-iters <#iterations>"
	//
	{ OPT_SPATIAL_TYPE,	_T("-spet"),     		SO_REQ_SEP	}, // "-spet <type#>"
	{ OPT_SPATIAL_TYPE,	_T("--spatialerr-type"),SO_REQ_SEP	}, // "--spatialerr-type <type#>"
	//
	{ OPT_NN_DEPTH,		_T("-nnd"),     		SO_REQ_SEP	}, // "-nnd <depth 1-16>"
	{ OPT_NN_DEPTH,		_T("--nn-depth"),		SO_REQ_SEP	}, // "--nn-depth <depth 1-16>"
	//
	{ OPT_NN_BREADTH,	_T("-nnb"),     		SO_REQ_SEP	}, // "-nnb <depth 0.0-1.0>"
	{ OPT_NN_BREADTH,	_T("--nn-breadth"),		SO_REQ_SEP	}, // "--nn-breadth <depth 0.0-1.0>"
	//
	{ OPT_NN_RATE,		_T("-nnr"),     		SO_REQ_SEP	}, // "-nnr <depth 0.0-1.0>"
	{ OPT_NN_RATE,		_T("--nn-rate"),		SO_REQ_SEP	}, // "--nn-rate <depth 0.0-1.0>"
	//
	{ OPT_PFM_GAMMA,	_T("-pfmg"),     		SO_REQ_SEP	}, // "-pfmg <float gamma>"
	{ OPT_PFM_GAMMA,	_T("--pfm-gamma"),   	SO_REQ_SEP	}, // "--pfm-gamma <float gamma>"
	//
	{ OPT_PFM_INFLUENCE,	_T("-pfmi"),     		SO_REQ_SEP	}, // "-pfmgi <float influence>"
	{ OPT_PFM_INFLUENCE,	_T("--pfm-influence"),  SO_REQ_SEP	}, // "--pfm-influence <float influence>"
	//
	{ OPT_PFM_SOURCE,	_T("-pfms"),     		SO_REQ_SEP	}, // "-pfms <maskfile.png>"
	{ OPT_PFM_SOURCE,	_T("--pfm-source"),  	SO_REQ_SEP	}, // "--pfm-source  <maskfile.png>"
	//
	{ OPT_PCT_SOURCE,	_T("-pct"),     		SO_REQ_SEP	}, // "-pfms <syncmapfile.png>"
	{ OPT_PCT_SOURCE,	_T("--pct-source"),  	SO_REQ_SEP	}, // "--pfm-source  <syncmapfile.png>"
	//
	{ OPT_DPT_SOURCE,	_T("-dpt"),     		SO_REQ_SEP	}, // "-pdpt <colmaskfile.png>"
	{ OPT_DPT_SOURCE,	_T("--dpt-source"),  	SO_REQ_SEP	}, // "--dpt-source  <colmaskfile.png>"
	//
	{ OPT_XSIZE,  		_T("-x"),     			SO_REQ_SEP	}, // "-x <size>"
	{ OPT_XSIZE,  		_T("--xsize"),     		SO_REQ_SEP	}, // "--xsize <size>"
	//
	{ OPT_YSIZE,  		_T("-y"),     			SO_REQ_SEP	}, // "-y <size>"
	{ OPT_YSIZE,  		_T("--ysize"),     		SO_REQ_SEP	}, // "--ysize <size>"
	//
	//{ OPT_PLANES,  		_T("-pc"),     			SO_REQ_SEP }, // "-pc <planecount>"
	//{ OPT_PLANES,  		_T("--plane-count"),     SO_REQ_SEP }, // "--plane-count <planecount>"
	//
	//{ OPT_MIN_COLOUR,  	_T("-cmin"),     		SO_REQ_SEP }, // "-cmin <first usable colour>"
	//{ OPT_MIN_COLOUR,  	_T("--colour-min"),     SO_REQ_SEP }, // "--colour-min <first usable colour>"
	//
	//{ OPT_MAX_COLOUR,  	_T("-cmax"),     		SO_REQ_SEP }, // "-cmin <last usable colour>"
	//{ OPT_MAX_COLOUR,  	_T("--colour-max"),     SO_REQ_SEP }, // "--colour-max <last usable colour>"
	//
	//{ OPT_RESERVE_COL,  _T("-cr"),     			SO_REQ_SEP }, // "-cr <reserve colour>"
	//{ OPT_RESERVE_COL,  _T("--colour-reserve"), SO_REQ_SEP }, // "--colour-reserve <reserve colour>"
	//
	//{ OPT_PALETTE_SIZE, _T("-pe"),     			SO_REQ_SEP }, // "-ps <#colours per palette>"
	//{ OPT_PALETTE_SIZE, _T("--pal-entries"), 	SO_REQ_SEP }, // "--pal-entries <entries in 1 palette>"
	//
	//{ OPT_PAL_STRIDE, 	_T("-ps"),     			SO_REQ_SEP }, // "-ps <#colours per scanline>"
	//{ OPT_PAL_STRIDE, 	_T("--pal-stride"), 	SO_REQ_SEP }, // "--pal-stride <entries per scanline>"
	//
	//{ OPT_PAL_PRELOAD, 	_T("-pp"),     			SO_REQ_SEP }, // "-pp <#colours preloaded>"
	//{ OPT_PAL_PRELOAD, 	_T("--pal-preload"), 	SO_REQ_SEP }, // "--pal-preload <#colours preloaded>"
	//
	{ OPT_OUTFILE,  	_T("-o"),				SO_REQ_SEP	}, 	// "-o <outfile>
	//
	SO_END_OF_OPTIONS
};

// --------------------------------------------------------------------

static void usage(int level)
{
	if (level >= 1) 
	{
		std::cout << version_string << std::endl;
		std::cout << "built on " << __DATE__ << std::endl;
	}
	
	//fprintf(stderr, "usage: pcs5 [options] [-o outfile.pcs] infile.jpg/png\n");
	//printf("press any key...\n");
	
	//////////////XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX 
	std::cout << "usage:" << std::endl;
	std::cout << " pcs5 [opts] [-o <outfile>] [infile(s)]" << std::endl;
	std::cout << std::endl;
	std::cout << " infile = filename or glob (*.ext)" << std::endl;
	std::cout << std::endl;
	std::cout << " -?, -h, --help		> show help/usage" << std::endl;
	std::cout << " -v, --verbose		> more output" << std::endl;
	std::cout << " -q, --quiet		> less output" << std::endl;
	std::cout << " -k, --wait-key		> wait key on exit" << std::endl;
	std::cout << " -dg, --diagnostics	> emit diagnostics" << std::endl;
	std::cout << std::endl;
	std::cout << " -f, --fields		> # image fields 1-2" << std::endl;
	std::cout << " -fm, --fieldmode		> # field mode 0,1,2" << std::endl;
	std::cout << " -m, --method		> reduction type 0-" << (ReduceMethod__NUM_-1) << std::endl;
	std::cout << " -lt, --lace-type	> i'lace method 0-" << (LaceType__NUM_-1) << std::endl;
	std::cout << " -dt, --dither-type	> input dither 0-" << (DitherType__NUM_-1) << std::endl;
	std::cout << " -dl, --dither-level	> dither strength 0.0-1.0" << std::endl;
	std::cout << " -et, --error-type	> output diffusion 0-" << (ErrorType__NUM_-1) << std::endl;
	std::cout << " -el, --error-level	> error feedback 0-100%" << std::endl;
	std::cout << " -ei, --error-iters	> FIELD error iterations" << std::endl;
	std::cout << " -spet, --spatialerr-type	> spatial error type 0-" << (SpatialType__NUM_-1) << std::endl;
	std::cout << " -nnd, --nn-depth     > neural net depth 1-16" << std::endl;
	std::cout << " -nnb, --nn-breadth   > neural net breadth 0.0-1.0" << std::endl;
	std::cout << " -nnr, --nn-rate      > neural net rate 0.0-1.0" << std::endl;
	std::cout << " -pfmg, --pfm-gamma	> PFM gamma" << std::endl;
	std::cout << " -pfmi			> PFM influence" << std::endl;
	std::cout << " -pfms, --pfm-source	> PFM source PNG" << std::endl;
	std::cout << std::endl;
	std::cout << " -pct, --pct-source	> palette conversion table (.CSV)" << std::endl;
	std::cout << " -dpt, --dpt-source	> dynamic point table (.CSV)" << std::endl;
	std::cout << std::endl;
	std::cout << " -cd, --colour-depth	> st/3, ste/4" << std::endl;
	std::cout << " -x, --xsize		> image width (320)" << std::endl;
	std::cout << " -y, --ysize		> image height (200)" << std::endl;
	std::cout << " -s, --storage		> output format 0-" << (StorageFmt__NUM_-1) << std::endl;
	std::cout << std::endl;
	std::cout << "press any key..." << std::endl;
				
	getchar();	
	exit(1);
}

// --------------------------------------------------------------------

static void GetArguments(int _argc, char **_argv, Arguments& _args)
{
    CSimpleOpt args(_argc, _argv, g_Options);
    std::string s_outfile;

    // while there are arguments left to process
    while (args.Next()) 
    {
		if (args.LastError() == SO_SUCCESS) 
		{
			if (args.OptionId() == OPT_HELP) 
			{
				usage(1);
			}
			else if (args.OptionId() == OPT_VERBOSE) 
			{
				_args.verbose = true;
			}
			else if (args.OptionId() == OPT_QUIET) 
			{
				_args.quiet = true;
			}
			else if (args.OptionId() == OPT_WAIT_KEY) 
			{
				_args.waitkey = true;
			}			
			else if (args.OptionId() == OPT_DIAGNOSTICS)
			{
				_args.diagnostics = true;
			}
			else if (args.OptionId() == OPT_XSIZE)
			{
				std::string s_xsize = args.OptionArg();
				int xsize = atoi(s_xsize.c_str());
				if (xsize < 2 || xsize > 32768)
				{
					printf("error - invalid x size: %s\n", s_xsize.c_str());
					exit(1);
				}
				
				_args.xsize = xsize;	
			}
			else if (args.OptionId() == OPT_YSIZE)
			{
				std::string s_ysize = args.OptionArg();
				int ysize = atoi(s_ysize.c_str());
				if (ysize < 2 || ysize > 32768)
				{
					printf("error - invalid y size: %s\n", s_ysize.c_str());
					exit(1);
				}
				
				_args.ysize = ysize;	
			}
			else if (args.OptionId() == OPT_FIELDS)
			{
				std::string s_f = args.OptionArg();
				int f = atoi(s_f.c_str());
				if (f < 0 || f > 2)
				{
					printf("error - invalid field count: %s\n", s_f.c_str());
					exit(1);
				}
				
				_args.fields = f;	
			}			
			else if (args.OptionId() == OPT_FIELDMODE)
			{
				std::string s_f = args.OptionArg();
				int f = atoi(s_f.c_str());
				if (f < 0 || f >= FieldMode__NUM_)
				{
					printf("error - invalid field mode: %s\n", s_f.c_str());
					exit(1);
				}
								
				_args.fieldmode = FieldMode(f);	
			}			
			else if (args.OptionId() == OPT_COL_DEPTH)
			{
				int depth = ColourDepth_DEFAULT;
				
				std::string s_m = args.OptionArg();
				std::transform(s_m.begin(), s_m.end(), s_m.begin(), mytolower);
					
				if ((s_m.find("ste") != std::string::npos) ||
					(s_m.compare("4096") == 0) ||
					(s_m.compare("12bit") == 0))
				{
					depth = ColourDepth_STE;	
				}
				else
				if ((s_m.compare("stfm") == 0) ||
					(s_m.compare("stf") == 0) ||
					(s_m.compare("st") == 0) ||
					(s_m.compare("512") == 0) ||
					(s_m.compare("9bit") == 0))
				{
					depth = ColourDepth_ST;	
				}
				else
				{						
					depth = atoi(s_m.c_str());
					if (depth < 3 || depth > 6)
					{
						printf("error - invalid native colour depth: %s\n", s_m.c_str());
						exit(1);
					}
				}
				_args.colourdepth = depth;	
			}
			else if (args.OptionId() == OPT_STORAGE)
			{
				std::string s_s = args.OptionArg();
				int s = atoi(s_s.c_str());
				if (s < 0 || s >= StorageFmt__NUM_)
				{
					printf("error - invalid storage format: %s\n", s_s.c_str());
					exit(1);
				}
				_args.storage = StorageFmt(s);	
			}
			else if (args.OptionId() == OPT_METHOD)
			{
				std::string s_m = args.OptionArg();
				int m = atoi(s_m.c_str());
				if (m < 0 || m >= ReduceMethod__NUM_)
				{
					printf("error - invalid method: %s\n", s_m.c_str());
					exit(1);
				}
				_args.method = ReduceMethod(m);	
			}
			else if (args.OptionId() == OPT_LACE_TYPE)
			{
				std::string s_m = args.OptionArg();
				int m = atoi(s_m.c_str());
				if (m < 0 || m >= LaceType__NUM_)
				{
					printf("error - invalid interlace type: %s\n", s_m.c_str());
					exit(1);
				}
				_args.lacetype = LaceType(m);	
			}
			else if (args.OptionId() == OPT_DITHER_TYPE)
			{
				std::string s_m = args.OptionArg();
				int m = atoi(s_m.c_str());
				if (m < 0 || m >= DitherType__NUM_)
				{
					printf("error - invalid dither type: %s\n", s_m.c_str());
					exit(1);
				}
				_args.dithertype = DitherType(m);	
			}	
			else if (args.OptionId() == OPT_DITHER_LEVEL)
			{
				std::string s_m = args.OptionArg();
				float m = atof(s_m.c_str());
				if (m < -8.0f || m > 8.0f)
				{
					printf("error - invalid dither level: %s\n", s_m.c_str());
					exit(1);
				}
				_args.ditherlevel = m;	
			}
			else if (args.OptionId() == OPT_ERROR_TYPE)
			{
				std::string s_e = args.OptionArg();
				int e = atoi(s_e.c_str());
				if (e < 0 || e >= ErrorType__NUM_)
				{
					printf("error - invalid error diffusion type: %s\n", s_e.c_str());
					exit(1);
				}
				_args.errortype = ErrorType(e);	
			}		
			else if (args.OptionId() == OPT_ERROR_LEVEL)
			{
				std::string s_el = args.OptionArg();
				float el = atof(s_el.c_str());
				if (el < 0.0f || el > 100.0f)
				{
					printf("error - invalid error level: %s%%\n", s_el.c_str());
					exit(1);
				}
				_args.errorlevel = xclamp<float>
				(
					(el / 100.0f), 0.0f, 1.0f
				);	
			}
			else if (args.OptionId() == OPT_ERROR_ITERS)
			{
				std::string s_i = args.OptionArg();
				int i = atoi(s_i.c_str());
				if (i < 1 || i >= 50)
				{
					printf("error - invalid error number of error iterations: %s\n", s_i.c_str());
					exit(1);
				}
				_args.erroriters = i;	
			}								
			else if (args.OptionId() == OPT_SPATIAL_TYPE)
			{
				std::string s_e = args.OptionArg();
				int e = atoi(s_e.c_str());
				if (e < 0 || e >= SpatialType__NUM_)
				{
					printf("error - invalid error diffusion type: %s\n", s_e.c_str());
					exit(1);
				}
				_args.spatialtype = SpatialType(e);	
			}		
			else if (args.OptionId() == OPT_PFM_GAMMA)
			{
				std::string s_m = args.OptionArg();
				float m = atof(s_m.c_str());
				if (m < 0.0 || m > 100.0)
				{
					printf("error - invalid PFM gamma: %s\n", s_m.c_str());
					exit(1);
				}
				_args.pfm_gamma = m;	
			}
			else if (args.OptionId() == OPT_PFM_INFLUENCE)
			{
				std::string s_m = args.OptionArg();
				float m = atof(s_m.c_str());
				if (m < 0.0 || m > 1.0)
				{
					printf("error - invalid PFM influence: %s (range 0.0 -> 1.0)\n", s_m.c_str());
					exit(1);
				}
				_args.pfm_influence = m;	
			}
			else if (args.OptionId() == OPT_NN_DEPTH)
			{
				std::string s_e = args.OptionArg();
				int e = atoi(s_e.c_str());
				if (e < 1 || e > 20)
				{
					printf("error - invalid NN depth: %s\n", s_e.c_str());
					exit(1);
				}
				_args.nn_depth = SpatialType(e);	
			}					
			else if (args.OptionId() == OPT_NN_BREADTH)
			{
				std::string s_m = args.OptionArg();
				float m = atof(s_m.c_str());
				if (m < 0.01 || m > 1.0)
				{
					printf("error - invalid NN breadth: %s (range 0.0 -> 1.0)\n", s_m.c_str());
					exit(1);
				}
				_args.nn_breadth = m;	
			}
			else if (args.OptionId() == OPT_NN_RATE)
			{
				std::string s_m = args.OptionArg();
				float m = atof(s_m.c_str());
				if (m < 0.0001 || m > 1.0)
				{
					printf("error - invalid NN rate: %s (range 0.0 -> 1.0)\n", s_m.c_str());
					exit(1);
				}
				_args.nn_rate = m;	
			}
			else if (args.OptionId() == OPT_PFM_SOURCE)
			{
				_args.pfm_source = args.OptionArg();
			}
			else if (args.OptionId() == OPT_PCT_SOURCE)
			{
				_args.pct_source = args.OptionArg();
			}										
			else if (args.OptionId() == OPT_DPT_SOURCE)
			{
				_args.dpt_source = args.OptionArg();
			}										
			else if (args.OptionId() == OPT_OUTFILE)
			{
				s_outfile = args.OptionArg();
			}
		}
		else 
		{
			printf("error - invalid argument: %s\n", args.OptionText());
			exit(1);
		}
    }

	// expand input filenames    
    CSimpleGlob glob(SG_GLOB_NODOT|SG_GLOB_NOCHECK);
    if (SG_SUCCESS != glob.Add(args.FileCount(), args.Files())) 
    {
        printf("error globbing files\n");
			exit(1);
    }
 
 	// check infile/outfile specification rules  
    if (glob.FileCount() == 0)
    {
		fprintf(stderr, "error - no input files specified!\n");
		printf("press any key...\n");
		getchar();	
		exit(1);
    }
	else if ((glob.FileCount() > 1) && !s_outfile.empty())
    {
		fprintf(stderr, "error - multiple inputs can't be converted to a single file! remove [-o <outfile>] from spec for batching\n");
		printf("press any key...\n");
		getchar();	
		exit(1);
    }
    else
    {
    	// collect input files
		for (int n = 0; n < glob.FileCount(); ++n) 
		{
		    std::string namein(glob.File(n));
		    std::string nameout(namein);
		    
		    // transform file extension to PCS
		    
			size_t p = nameout.find_last_of(".");
			if (p != std::string::npos)
				nameout = nameout.substr(0, p);
			nameout.append(".PCS");			
			
		    _args.tasks.push_back(task(namein.c_str(), nameout.c_str()));
		}
		
		// if only one input file, specifying the outfile directly is ok
		if (!s_outfile.empty())
		{
			_args.tasks[0].outfile_ = s_outfile;
		}
 	}			

	// read user PCT file, if any 
/*
	// convert old DPT mask values into new format
	int i = 0;	
	for (int y = 0; y < 40; ++y)
	{
		for (int x = 0; x < 8; ++x)
		{
			unsigned short v = DPT_PCS_original[i] >> 8;
	
			if (v == 16)
				printf("%04x, ", 0);
			else
				printf("%04x, ", (int)(1 << v));
				
			++i;
		}
		printf("\n");
	}
*/	
	if (_args.pct_source.empty())
	{
		if (_args.xsize == 320)
		{
			// use builtin table for PCS
			_args.aPCT = PCT_PCS;
		}	
		else
		{
			// xsize has been specified but without a display profile - most likely to 
			// conveniently leverage the image handling parts of the tool without generating 
			// any display-specific output...
			std::cout << "warning - specified xsize independently of PCT - image processing will continue but multipalette reduction steps will be skipped..." << std::endl;

			_args.aPCT = 0;
			_args.uPCT = 0;

			//_args.uPCT = new byte_t[_args.xsize*16];
			//memset(_args.uPCT, 0, _args.xsize*16);
			//_args.aPCT = _args.uPCT;
		}	
	}
	else
	{
		_args.uPCT = new byte_t[_args.xsize*16];
		memset(_args.uPCT, 0, _args.xsize*16);
		_args.aPCT = _args.uPCT;
		
		std::ifstream pct(_args.pct_source.c_str());
		
		if (!_args.quiet)
			std::cout << "reading user-defined PCT (pixel conversion / timing table)..." << std::endl;

		int validrows = 0;					
		int validentries = 0;					
		bool firstrow = true;
		int valid_check = _args.xsize;
		for (CSVIterator loop(pct); loop != CSVIterator(); ++loop)
		{
			const CSVRow& r = (*loop);
			size_t s = r.size();
			if (s > 0)
			{
				if (s > 1 && s <= 16)
				{
					if (validrows >= valid_check)
					{
						fprintf(stderr, "error - invalid number of rows in PCT file!\n");
						printf("press any key...\n");
						getchar();	
						exit(1);				
					}
					
					if (firstrow && (s == 5))
					{
						// read profile data
						// <width>, <ncolours>, <npalettes>, <nskips>, <firstline> 
						int column_value = 0;

						std::cout << "reading custom profile: ";

						// <width>		
						std::string colval = r[0];
						std::transform(colval.begin(), colval.end(), colval.begin(), sanitise_csv_arg);
						column_value = atoi(colval.c_str());
						if (column_value >= 16 && column_value <= 2048)
						{	
							_args.xsize = column_value;
							
							// reallocate
							delete[] _args.uPCT;
							_args.uPCT = new byte_t[_args.xsize*16];
							_args.aPCT = _args.uPCT;
							
							valid_check = _args.xsize;
							std::cout << "width=" << column_value << ", ";
						}
						else
						{
							fprintf(stderr, "error - invalid <width> field (%d) in profile row of PCT CSV file!\n", column_value);
							printf("press any key...\n");
							getchar();	
							exit(1);				
						}
						
						// <ncolours>		
						colval = r[1];
						std::transform(colval.begin(), colval.end(), colval.begin(), sanitise_csv_arg);
						column_value = atoi(colval.c_str());
						if (column_value >= 3 && column_value <= 256)
						{	
							_args.colours = column_value;
							std::cout << "ncolours=" << column_value << ", ";
						}
						else
						{
							fprintf(stderr, "error - invalid <ncolours> field (%d) in profile row of PCT CSV file!\n", column_value);
							printf("press any key...\n");
							getchar();	
							exit(1);				
						}							
						
						// <npalettes>		
						colval = r[2];
						std::transform(colval.begin(), colval.end(), colval.begin(), sanitise_csv_arg);
						column_value = atoi(colval.c_str());
						if (column_value >= 1 && column_value <= 8)
						{	
							_args.palettes = column_value;
							std::cout << "npalettes=" << column_value << ", ";
						}
						else
						{
							fprintf(stderr, "error - invalid <npalettes> field (%d) in profile row of PCT CSV file!\n", column_value);
							printf("press any key...\n");
							getchar();	
							exit(1);				
						}
						
						// <nskips>		
						colval = r[3];
						std::transform(colval.begin(), colval.end(), colval.begin(), sanitise_csv_arg);
						column_value = atoi(colval.c_str());
						if (column_value >= 0 && column_value < (_args.ysize * _args.palettes))
						{	
							_args.palette_skip = column_value;
							std::cout << "nskips=" << column_value << ", ";
						}
						else
						{
							fprintf(stderr, "error - invalid <nskips> field (%d) in profile row of PCT CSV file!\n", column_value);
							printf("press any key...\n");
							getchar();	
							exit(1);				
						}
						
						// <firstline>		
						colval = r[4];
						std::transform(colval.begin(), colval.end(), colval.begin(), sanitise_csv_arg);
						column_value = atoi(colval.c_str());
						if (column_value >= 0 && column_value < 512)
						{	
							_args.firstline = column_value;
							std::cout << "firstline=" << column_value << std::endl;
						}
						else
						{
							fprintf(stderr, "error - invalid <firstline> field (%d) in profile row of PCT CSV file!\n", column_value);
							printf("press any key...\n");
							getchar();	
							exit(1);				
						}																									
					}
					else
					{
						// read timing data
					
					for (size_t c = 0; c < s; ++c)
					{
						std::string colval = r[c];
						std::transform(colval.begin(), colval.end(), colval.begin(), sanitise_csv_arg);
						int column_value = atoi(colval.c_str());
						
						if (column_value >= _args.palettes)
						{
							fprintf(stderr, "error - invalid palette index (%d) in PCT table >= palettes per line (%d)!\n", column_value, _args.palettes);
							printf("press any key...\n");
							getchar();	
							exit(1);				
						}							
												
						_args.uPCT[(validrows * 16) + c] = (unsigned char)(unsigned int)column_value;
						validentries++;

#ifdef _DEBUG
						if (_args.verbose)
						{
							if (column_value >= 0)
								std::cout << " ";	
							std::cout << column_value;
							if (c != s-1)
								std::cout << ", ";
						}
#endif						
					}
#ifdef _DEBUG
					if (_args.verbose)
						std::cout << std::endl;
#endif
					validrows++;
				
					}
					firstrow = false;
				}
				else
				{
					// syntax error in PCT
					fprintf(stderr, "error - invalid number of columns in PCT file!\n");
					printf("press any key...\n");
					getchar();	
					exit(1);
				}
			}
		} 
		if (_args.verbose)	
			std::cout << validrows << " rows read from CSV" << std::endl;
	}
 	
	// read user DPT file, if any 
	
	if (_args.dpt_source.empty())
	{
		if (_args.aPCT == PCT_PCS)
		{
			// default is to use builtin DPT table for PCS
			_args.aDPT = DPT_PCS;
		}
		else
		{
			// otherwise, don't use any DPT table!
			std::cout << "warning - no DPT specified - if image displays vertical bands you may need to mask colours under changing pixels" << std::endl;
		}
	}
	else
	{
		_args.uDPT = new word_t[_args.xsize];
		_args.aDPT = _args.uDPT;
		
		std::ifstream dpt(_args.dpt_source.c_str());
		
		if (!_args.quiet)
			std::cout << "reading user-defined DPT (dynamic point table / colour masks)..." << std::endl;

		int validrows = 0;					
		int validentries = 0;					
		for (CSVIterator loop(dpt); loop != CSVIterator(); ++loop)
		{
			const CSVRow& r = (*loop);
			size_t s = r.size();
			if (s > 0)
			{
				if (s >= 1 && s <= 768)
				{
					for (size_t c = 0; c < s; ++c)
					{
						if (validentries >= _args.xsize)
						{
							fprintf(stderr, "error - invalid number of entries in DPT file!\n");
							printf("press any key...\n");
							getchar();	
							exit(1);				
						}
						
						std::string colval = r[c];
						std::transform(colval.begin(), colval.end(), colval.begin(), sanitise_csv_arg);
						int column_value = -1;
						sscanf(colval.c_str(), "%x", &column_value);
																		
						_args.uDPT[validentries++] = (unsigned short)(unsigned int)column_value;

#ifdef _DEBUG
						if (_args.verbose)
						{
							//std::cout << "[" << colval.c_str() << " : " << column_value << "]";
							
							fprintf(stdout, "%04x", column_value);
							if (c != s-1)
								std::cout << ", ";
						}
#endif
					}
#ifdef _DEBUG
					if (_args.verbose)
						std::cout << std::endl;
#endif							
					validrows++;
				}
				else
				{
					// syntax error in DPT
					fprintf(stderr, "error - invalid number of columns in DPT file!\n");
					printf("press any key...\n");
					getchar();	
					exit(1);
				}
			}
		} 
		
		if (validentries != _args.xsize)
		{
			// syntax error in DPT
			fprintf(stderr, "error - too few entries (%d) in DPT file! should correspond to display width (%d).\n", validentries, _args.xsize);
			printf("press any key...\n");
			getchar();	
			exit(1);
		}
		else
		{			
			if (_args.verbose)	
				std::cout << validentries << " entries read from CSV" << std::endl;
		}
	}
	 	
 	// override interlace control in single-field mode
 	if (_args.fields == 1)
 		_args.lacetype = LaceType_OFF;

	if ((_args.fieldmode == FieldMode_SharedPalette) && (_args.fields == 1))
	{
		fprintf(stderr, "error - 'shared palette' field mode requires fields==2 for colour reduction\n");
		printf("press any key...\n");
		getchar();	
		exit(1);
	}

	if ((_args.fieldmode == FieldMode_SharedBitmap) && (_args.fields == 2))
	{
		fprintf(stderr, "error - 'shared bitmap' field mode requires fields==1 for colour reduction\n");
		printf("press any key...\n");
		getchar();	
		exit(1);
	}
	
	if (_args.fieldmode == FieldMode_SharedBitmap)
	{
		// sharing bitmap implies increased palette colour depth in order to generate 2 palette fields from one reduction
		// note: this requires special care on STE for intermediate storage of RGB in 12bit palette form, as 15bit colour will be generated during reduction...
		_args.colourdepth++;

		if (_args.verbose)	
			std::cout << "extending bitdepth for shared bitmap / dual palette image" << std::endl;
		
	}

	// don't accept more than one error iteration unless we're in a field-error propagation mode 		
 	if (
 		((_args.errortype != ErrorType_FIELD) /*&& 
 		 (_args.errortype != ErrorType_MULTIMODE)*/) /*|| 
 		(_args.fields < 2)*/
 	)
 		_args.erroriters = 1;
 	
}

bool process_file(const char* const _name, const char* _outfile);
	
Arguments g_allargs;

int main(int argc, char **argv)
{
	GetArguments(argc, argv, g_allargs);

	if (!g_allargs.quiet)
		printf("%s\n", version_string);

	if (g_allargs.verbose)
	{
		switch (g_allargs.colourdepth)
		{
		case ColourDepth_ST:
			printf("colour depth: ST palette (9bit)\n");
			break;
		case ColourDepth_STE:
			printf("colour depth: STE palette (12bit)\n");
			break;
		case ColourDepth_STE_X:
			printf("colour depth: STE dual palette (15bit)\n");
			break;
		case ColourDepth_FALCON:
			printf("colour depth: Falcon030 (18bit) - not yet properly supported!\n");
			exit(1);			
			break;
		default:
			printf("no obvious use case for %d-bit colour - are you sure?\n", g_allargs.colourdepth * 3);
			break;
		}
	} 
	   	
	if (g_allargs.verbose)
		for (int t = 0; t < g_allargs.tasks.size(); ++t)
			printf("task: [%s->%s]\n", 
				g_allargs.tasks[t].infile_.c_str(), 
				g_allargs.tasks[t].outfile_.c_str()
			);


	bool success = false;

	if (!g_allargs.tasks.empty())
	{
		success = true;
		for (int t = 0; t < g_allargs.tasks.size(); ++t)
		{
			success &= process_file(g_allargs.tasks[t].infile_.c_str(), g_allargs.tasks[t].outfile_.c_str());
		}
	}
	
	if (success)
	{
		if (!g_allargs.quiet)		
			printf("complete!\n"); 
			
		if (g_allargs.waitkey) 
			getchar();
	}
	else
	{
		printf("error - press any key...\n"); getchar();
	}

	return 0;
}
